#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <dlfcn.h>
#include <pthread.h>
#include <stdint.h>

#include <iomanip>
#include <iostream>
#include <fstream>

#include <math.h>

using namespace std;

#define INITGUID

#include "lcomp/e2010cmd.h"
#include "lcomp/ifc_ldev.h"
#include <errno.h>

#include <signal.h>
#include <unistd.h>

typedef IDaqLDevice* (*CREATEFUNCPTR)(ULONG Slot);

/* ----------------    Параметры теста  ---------------------------  */
#define ADC_FREQ_KHZ  10000 /* частота АЦП в КГц */
#define USE_TEST_COUNTER /* использование тестовой последовательности с ее проверкой вместо
                            собственно данных с АЦП */
//#define USE_DIG_CHANNEL  /* использование цифровых каналов в лог. таблице АЦП */
//#define USE_DEVBUF_OVF_CHECK /* периодическая проверка состояния буфера при сборе (по завершению сбора выполняется всегда) */
#define STOP_ON_DATA_ERROR /* останов сбора при первой ошибки с доп. приемом половины сохраняемого буфера */

/* размер выделяемого буфера в драйвере для сбора данных в словах */
#define LCOMP_BUF_SIZE (32 * 1024 * 1024)
#define IRQ_STEP       (1024 * 1024) /* размер одного обмена в словах */
/* ----------------------------------------------------------------- */

#define PAGES_CNT (LCOMP_BUF_SIZE/IRQ_STEP)
#if LCOMP_BUF_SIZE % IRQ_STEP != 0
    #error lcomp buf size must be integer mul of irq_step
#endif

#define SAVE_BUF_SIZE  (1024*1024)
#define SAVE_ERRS_CNT  64

#ifdef USE_TEST_COUNTER
    #define TEST_COUNTER_MODULE 65535
    /* для цифровых каналов тестовое слово не передается
       в интерфейс. в пример используется попеременно
       цифровые и аналогвые каналы и из-за отстутсвия
       слов цифровых счетчик будет инкрементироваться
       на 2, а не 1 */
    #ifdef USE_DIG_CHANNEL
        #define TEST_COUNTER_INC 2
    #else
        #define TEST_COUNTER_INC 1
    #endif

    /* для частот 5000, 6000 и 7500 идет некорректная
       тестовая последовательность со сдвигом на байт */
    #if ((ADC_FREQ_KHZ < 8750) && (ADC_FREQ_KHZ > 4650))
        #define TEST_COUNTER_BYTE_SHIFT
    #endif
#endif


#pragma pack(1)

#define E2010B_DATA_STATUS_BUF_OVF     0x01 /* признак переполнения внутреннего буфера АЦП */
#define E2010B_DATA_STATUS_CHS_OVL     0x08  /* общий признак перегрузки входа для всех каналов АЦП */
#define E2010B_DATA_STATUS_CH_OVL(ch)  (0x10 << ch) /* признак перегрузки входа указанного канала АЦП */

/* статистика о состоянии буфера модуля, читаемая по адресу DATA_STATE_ADDRESS */
typedef struct {
    uint8_t  status; /* набор флагов из E2010B_DATA_STATUS_xxx */
    uint32_t cur_buffer_filling; /* текущая заполненность вутренного буфера модуля в словах */
    uint32_t max_buf_filling;    /* максимальная заполненность бфера модуля в словах */
    uint32_t buf_size; /* размер буфера модуля в словах */
} t_e2010b_data_state;
#pragma pack()



unsigned short *p;
unsigned int *pp;

/* признак необходимости завершить сбор данных */
static int f_out = 0;
#ifndef _WIN32
/* Обработчик сигнала завершения для Linux */
static void f_abort_handler(int sig) {
    f_out = 1;
}
#endif

static uint16_t tmp_buf[SAVE_BUF_SIZE];
static unsigned tmp_buf_pos = 0;
static unsigned tmp_buf_size = 0;

struct t_err_info {
    unsigned tmp_buf_pos; /* позиция в tmb_buf слова с первой ошибкой */
    unsigned lcomp_buf_pos; /* позиция слова с ошикой в буере pp из драйвера lcomp */
    unsigned proc_buf_pos; /* позиция слова в текущим обрабатываемом блоке данных */
    unsigned proc_buf_len;
    uint64_t start_block_dtime; /* время, прошедшее с момента начала обработки прошлого
                                   блока до начала обработки блока с первой ошибкой в мкс */
    uint64_t err_point_dtime; /* время от начала обработки блока до нахождения
                                 первой ошибки в мкс */
    uint64_t word_num; /* номер слова с начала сбора */

    uint16_t rcv_wrd;
    uint16_t exp_wrd;
};


static struct {
    t_err_info errs[SAVE_ERRS_CNT];
    uint64_t errs_cnt;

    uint64_t total_samples;
    uint64_t total_time;
    uint64_t max_block_time;
    unsigned max_block_len;
    unsigned first_block_len;


    struct {
        int      ov_cnt;   /* признак обнаружения переполнения */
        uint32_t max_words; /* макс. заполненность буфера в словах */
        uint32_t buf_size;
        uint64_t max_req_dtime; /* макс. время выполнения запроса чтения состояния буфера в мкс */
    } module_buf;


} buf_stat;






static uint64_t timespec_diff_mks(timespec start, timespec end) {
    timespec temp;
    if ((end.tv_nsec-start.tv_nsec)<0) {
        temp.tv_sec = end.tv_sec-start.tv_sec-1;
        temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
    } else {
        temp.tv_sec = end.tv_sec-start.tv_sec;
        temp.tv_nsec = end.tv_nsec-start.tv_nsec;
    }
    return temp.tv_sec * 1000000 + temp.tv_nsec/1000;
}


static void process_data(unsigned int from, unsigned to) {
    static unsigned tmp_buf_stop_pos;
    static int      tmp_buf_stop_req = 0;
#ifdef USE_TEST_COUNTER
    static uint64_t tst_cur_cntr = 0;
#endif

    unsigned proc_len = to - from;
    if (proc_len > 0) {
        static int tmp_buf_prev_time_valid = 0;
        static timespec tmp_buf_prev_time;
        uint64_t block_time = 0;

        /* обновление статситки по времени между блоками */
        timespec block_start_time;
        clock_gettime(CLOCK_MONOTONIC, &block_start_time);
        if (tmp_buf_prev_time_valid) {
            block_time = timespec_diff_mks(tmp_buf_prev_time, block_start_time);
            if (block_time > buf_stat.max_block_time)
                buf_stat.max_block_time = block_time;
            buf_stat.total_time += block_time;
        } else {
            tmp_buf_prev_time_valid = 1;
            buf_stat.first_block_len = proc_len;
        }
        tmp_buf_prev_time = block_start_time;

        if (proc_len > buf_stat.max_block_len)
            buf_stat.max_block_len = proc_len;

        /*if (buf_stat.errs_cnt == 0) {
            cout << "proc "  << block_time << " " << proc_len << endl;
        }*/

        for (unsigned int idx = from; idx < to; idx++) {
            unsigned short wrd = p[idx];
            tmp_buf[tmp_buf_pos] = wrd;
#ifdef USE_TEST_COUNTER
#ifdef TEST_COUNTER_BYTE_SHIFT
            uint16_t exp_wrd = buf_stat.total_samples == 0 ? 0 :
                               (((tst_cur_cntr % TEST_COUNTER_MODULE) & 0xFF) << 8) |
                               ((((tst_cur_cntr + 1) % TEST_COUNTER_MODULE) >> 8) & 0xFF);
#else
            uint16_t exp_wrd = (tst_cur_cntr % TEST_COUNTER_MODULE);
#endif
            if (wrd != exp_wrd) {
                clock_gettime(CLOCK_MONOTONIC, &block_start_time);
                if (buf_stat.errs_cnt < SAVE_ERRS_CNT) {
                    timespec err_time;
                    clock_gettime(CLOCK_MONOTONIC, &err_time);
                    t_err_info *perr_info = &buf_stat.errs[buf_stat.errs_cnt];

                    perr_info->tmp_buf_pos = tmp_buf_pos;
                    perr_info->lcomp_buf_pos = idx;
                    perr_info->proc_buf_pos = idx - from;
                    perr_info->proc_buf_len = proc_len;
                    perr_info->start_block_dtime = block_time;
                    perr_info->err_point_dtime = timespec_diff_mks(block_start_time, err_time);
                    perr_info->word_num = buf_stat.total_samples;
                    perr_info->rcv_wrd = wrd;
                    perr_info->exp_wrd = exp_wrd;

                    /* при первой ошибке планируем завершение приема через пол сохраняемого буфера от возникшей ошибки,
                       чтобы сбойное слово было посередине кольцевого буфера и иметь контекст как до,
                       так и после ошики */
                    if (buf_stat.errs_cnt == 0) {
                        tmp_buf_stop_pos = (tmp_buf_pos + SAVE_BUF_SIZE/2) % SAVE_BUF_SIZE;
#ifdef STOP_ON_DATA_ERROR
                        tmp_buf_stop_req = 1;
#endif
                    }
                }
                buf_stat.errs_cnt++;
                tst_cur_cntr = wrd;
            }
            tst_cur_cntr+=TEST_COUNTER_INC;
#endif
            if (++tmp_buf_pos == SAVE_BUF_SIZE) {
                tmp_buf_pos = 0;                
                cout <<  (buf_stat.errs_cnt == 0 ? "." : "!") << flush;
            }

            if (tmp_buf_size < SAVE_BUF_SIZE)
                tmp_buf_size++;
            /* если был запланирован останов по ошибке и дошли до запланированной
               позиции останов, то завершаем сор и обработку */
            if (tmp_buf_stop_req && (tmp_buf_pos == tmp_buf_stop_pos)) {
                f_out = 1;
                break;
            }

            buf_stat.total_samples++;
        }
    }
}

int check_data_state(IDaqLDevice *pI) {
    /* Читаем из памяти модуля статистику модуля по передаче данных.
       используем inmbyte для чтения данных нужного размера */
    t_e2010b_data_state dstate;
    timespec req_start_time;
    clock_gettime(CLOCK_MONOTONIC, &req_start_time);
    memset(&dstate, 0, sizeof(dstate));

    int err = pI->inmbyte(DATA_STATE_ADDRESS, reinterpret_cast<PUCHAR>(&dstate), sizeof(dstate));
    if (err != L_SUCCESS) {
        cout << endl << "get data state error " << err << endl;
    } else {
        uint64_t req_exec_dtime;
        timespec req_finish_time;
        clock_gettime(CLOCK_MONOTONIC, &req_finish_time);
        req_exec_dtime = timespec_diff_mks(req_start_time, req_finish_time);
        if (req_exec_dtime > buf_stat.module_buf.max_req_dtime) {
            buf_stat.module_buf.max_req_dtime = req_exec_dtime;
        }

        if (dstate.status & E2010B_DATA_STATUS_BUF_OVF) {
            buf_stat.module_buf.ov_cnt++;
            err = -10;
            cout << endl << "module buffer overflow detected!" << endl;
        }

        if (dstate.max_buf_filling > buf_stat.module_buf.max_words) {
            buf_stat.module_buf.max_words = dstate.max_buf_filling;
        }
        buf_stat.module_buf.buf_size = dstate.buf_size;
    }
    return err;
}

void *thread_func(void *arg) {
    unsigned int L_buf=0, L_buf_old=0;
    unsigned int L_buf_max = IRQ_STEP * PAGES_CNT;
    IDaqLDevice *pI = static_cast<IDaqLDevice *>(arg);

     cout << "data thread started..." << endl;

    while (!f_out) {
        usleep(1000);

        L_buf=*pp;
        L_buf-=L_buf%4; // выравниваем по кадру (в кадре 4 канала)

        if(L_buf != L_buf_old)  {
            if(L_buf_old < L_buf) {
                process_data(L_buf_old, L_buf);
            } else if(L_buf_old>L_buf) {
                // был "перескок" через максимум
                process_data(L_buf_old, L_buf_max);
                process_data(0, L_buf);
            }

            L_buf_old = L_buf;

#ifdef USE_DEVBUF_OVF_CHECK
            if (check_data_state(pI) != L_SUCCESS) {
                f_out = 1;
            }
#endif
        }
    }//for(;;)

    check_data_state(pI);

    cout << endl << "data thread finished..." << endl;

    return nullptr;
}


//Att. for board slot numbers!!!!

int main(int argc, char **argv) {
    int err = L_SUCCESS;
    void *handle = NULL;
    CREATEFUNCPTR CreateInstance = NULL;
    IDaqLDevice* pI = NULL;

    PLATA_DESCR_U2 pd;
    SLOT_PAR sl;
    ADC_PAR adcPar;
    ULONG size;

    const char *fw_filename = NULL;
    int dev_slot = 0;
#ifndef _WIN32
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    /* В ОС Linux устанавливаем свой обработчик на сигнал закрытия,
       чтобы завершить сбор корректно */
    sa.sa_handler = f_abort_handler;
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGABRT, &sa, NULL);
#endif

    if (argc > 2) {
        fw_filename = argv[2];
    }
    if (argc > 1) {
        dev_slot = atoi(argv[1]);
    }


    if (err == L_SUCCESS) {
        handle = dlopen("liblcomp.so", RTLD_LAZY);
        if(!handle) {
            cout << "error open dll!! " << dlerror() << endl;
            err = -2;
        }
    }

    if (err == L_SUCCESS) {
        CreateInstance = (CREATEFUNCPTR) dlsym(handle,"CreateInstance");
        if (!CreateInstance) {
            cout << "load symbol CreateInstance error " << dlerror() << endl;
            err = -3;
        }
    }


    if (err == L_SUCCESS) {
        LUnknown* pIUnknown = CreateInstance(dev_slot);
        if(pIUnknown == NULL) {
            cout << "Call CreateInstance failed " << errno << endl;
            err = -4;
        } else {
            cout << "Get IDaqLDevice interface" << endl;

            HRESULT hr = pIUnknown->QueryInterface(IID_ILDEV,(void**)&pI);
            if (hr != S_OK) {
                cout << "Get IDaqLDevice failed" << endl;
                err = -5;
            } else {
                printf("IDaqLDevice get success \n");
            }
        }

        if (pIUnknown) {
            pIUnknown->Release();
            cout << "Free IUnknown" << endl;
        }
    }

    if (err == L_SUCCESS) {
        if (pI->OpenLDevice() == INVALID_HANDLE_VALUE) {
            cout << "open device error" << endl;
            err = -6;
        } else {
            err = pI->GetSlotParam(&sl);
            if (err != L_SUCCESS) {
                cout << "get slot param error: " << err << endl;
            }
        }
    }

    if (err == L_SUCCESS) {
        err = pI->LoadBios(fw_filename);
        if (err) {
            cout << "Load bios failed: " << err << endl;
        }
    }

    if (err == L_SUCCESS) {
        err =  pI->PlataTest();
        if (err) {
            cout << "Plate test failed: " << err << endl;
        }
    }


    if (!err) {
        cout << endl << "Read FLASH" << endl;

        err = pI->ReadPlataDescr(&pd);
        if (err) {
            cout << "Read plata descr failed: " << err;
        } else {
            if (sl.BoardType == E2010B) {
                cout << "SerNum       " << pd.t6.SerNum << endl;
                cout << "BrdName      " << pd.t6.BrdName << endl;
                cout << "Rev          " << pd.t6.Rev << endl;
                cout << "DspType      " << pd.t6.DspType << endl;
                cout << "IsDacPresent " << pd.t6.IsDacPresent << endl;
                cout << "Quartz       " << dec << pd.t6.Quartz << endl;
            } else {
                cout << "unsupported board type: " << sl.BoardType;
                err = -6;
            }
        }
    }

    if (err == L_SUCCESS) {
        size = IRQ_STEP * PAGES_CNT;
        err = pI->RequestBufferStream(&size);
        if (err != L_SUCCESS) {
            cout << "request buffer stream error " << err << endl;
        }
    }

    if (err == L_SUCCESS) {
        if (sl.BoardType == E2010B) {
            adcPar.t2.s_Type = L_ADC_PARAM;
            adcPar.t2.AutoInit = 1;
            adcPar.t2.dRate = ADC_FREQ_KHZ;
            adcPar.t2.dKadr = 0.000;

            adcPar.t2.SynchroType = INT_START_TRANS;
            adcPar.t2.SynchroSrc = INT_CLK_TRANS;
            adcPar.t2.AdcIMask = SIG_0|SIG_1|SIG_2|SIG_3;
            adcPar.t2.DigRate = 0;

            adcPar.t2.NCh = 4;
#ifdef USE_DIG_CHANNEL
            adcPar.t2.Chn[0] = 2; // BNC3 (координата Y)
            adcPar.t2.Chn[1] = 6; // SYNC
            adcPar.t2.Chn[2] = 1; // BNC2 (координата X)
            adcPar.t2.Chn[3] = 6; // SYNC
#else
            adcPar.t2.Chn[0] = 2; // BNC3 (координата Y)
            adcPar.t2.Chn[1] = 0; //
            adcPar.t2.Chn[2] = 1; // BNC2 (координата X)
            adcPar.t2.Chn[3] = 3; //

#endif

            adcPar.t2.FIFO = IRQ_STEP;
            adcPar.t2.IrqStep = IRQ_STEP;
            adcPar.t2.Pages = PAGES_CNT;
            adcPar.t2.IrqEna = 1;
            adcPar.t2.AdcEna = 1;

            // extra sync mode
            adcPar.t2.StartCnt = 0;
            adcPar.t2.StopCnt = 0;
            adcPar.t2.DM_Ena = 0;
            adcPar.t2.SynchroMode = 0;
            adcPar.t2.AdPorog = 0;


            err = pI->FillDAQparameters(&adcPar.t2);
            if (err != L_SUCCESS) {
                cout << "fill daq parameters error " << err << endl;
            }

            if (err == L_SUCCESS) {
                pI->SetParametersStream(&adcPar.t2, &size, (void **)&p, (void **)&pp,L_STREAM_ADC);
                if (err != L_SUCCESS) {
                    cout << "set daq parameters error " << err << endl;
                }
            }

            if (err == L_SUCCESS) {
                cout << "Buffer size(word): " << size << endl;
                cout << "Pages:             " << adcPar.t2.Pages << endl;
                cout << "IrqStep:           " << adcPar.t2.IrqStep << endl;
                cout << "FIFO:              " << adcPar.t2.FIFO << endl;
                cout << "Rate:              " << adcPar.t2.dRate << " KHz" << endl;
                cout << "Kadr:              " << adcPar.t2.dKadr << " ms" << endl;

                cout << "Use dig channels   " <<
#ifdef USE_DIG_CHANNEL
                     "yes" << endl;
#else
                    "no" << endl;
#endif
                cout << "Use test sequence  " <<
#ifdef USE_TEST_COUNTER
                     "yes" << endl;
#else
                    "no" << endl;
#endif
                cout << "Dev buf ovf check  " <<
#ifdef USE_DEVBUF_OVF_CHECK
                     "yes" << endl;
#else
                    "no" << endl;
#endif
            }
        }

        if (err == L_SUCCESS) {
            err = pI->EnableCorrection();
            if (err != L_SUCCESS)
                cout << "Enabled correction error" << err << endl;
        }

        if (err == L_SUCCESS) {
            USHORT tst_wrd =
#ifdef USE_TEST_COUNTER
                    1;
#else
                    0;
#endif
            err = pI->outmword(SEL_TEST_MODE, &tst_wrd);
            if (err != L_SUCCESS)
                cout << "set test mode error" << err << endl;
        }


        if (err == L_SUCCESS) {
            err = pI->InitStartLDevice();
            if (err != L_SUCCESS)
                cout << "init start error" << err << endl;
        }

        if (err == L_SUCCESS) {
            pthread_t thread1;
            if (err == L_SUCCESS) {
                pthread_create(&thread1, NULL, thread_func, pI);
                err = pI->StartLDevice();
                if (err != L_SUCCESS) {
                    cout << "start device error" << err << endl;
                }
            }

            if (err == L_SUCCESS) {
                cout << "Data acquisition started. Press CTRL+C for stop..." << endl;
                while (!f_out) {
                    usleep(40000);
                }
            }

            pthread_join(thread1,NULL);
            pI->StopLDevice();

            /* вывод общией статистики по сбору */
            cout << "Data acqusition finished..."  << endl;
            cout << " Total samples processed: " << buf_stat.total_samples  << endl;
            cout << " Total time (mks):        " << buf_stat.total_time  << endl;
            cout << " Max block len:           " << buf_stat.max_block_len  << endl;
            cout << " Max block time (mks):    " << buf_stat.max_block_time << endl;
            if (buf_stat.total_time > 0) {
                cout << " Avg receive freq         " <<
                        (double)((uint64_t)1000000 *
                                 (buf_stat.total_samples - buf_stat.first_block_len)
                                 / buf_stat.total_time)/1000 << " KHz" << endl;
            }


            cout << " Module buffer state:" << endl;
            cout << "   overflow cntr:         " << buf_stat.module_buf.ov_cnt << endl;
            cout << "   max words:             " << buf_stat.module_buf.max_words << " of " << buf_stat.module_buf.buf_size << endl;
            cout << "   max request dtime:     " << buf_stat.module_buf.max_req_dtime << endl;


            cout << " Total errors:            " << buf_stat.errs_cnt << endl;

            if (buf_stat.errs_cnt > 0) {
                for (uint64_t err_idx = 0; (err_idx < buf_stat.errs_cnt) && (err_idx < SAVE_ERRS_CNT); err_idx++) {
                    cout << " Error " << err_idx + 1 << ":" << endl;
                    const t_err_info *perr_info = &buf_stat.errs[err_idx];
                    if (err_idx != 0) {
                        cout << "   words from prev err    " << perr_info->word_num - buf_stat.errs[err_idx-1].word_num << endl;
                    }
                    cout << "   lcomp buf pos:         " << perr_info->lcomp_buf_pos << endl;
                    cout << "   proc block pos:        " << perr_info->proc_buf_pos  << endl;
                    cout << "   proc block len:        " << perr_info->proc_buf_len  << endl;
                    cout << "   block start dtime:     " << perr_info->start_block_dtime  << endl;
                    cout << "   error point dtime:     " << perr_info->err_point_dtime  << endl;
                    cout << "   receive word:          " << perr_info->rcv_wrd  << endl;
                    cout << "   expected word:         " << perr_info->exp_wrd  << endl;
                }


                cout << "dump last data buffer to file..." << endl;
                /* сохранение кольцевого буфера в текстовый файл */
                ofstream dumpfile;
                dumpfile.open ("dump.txt");

                unsigned cur_buf_pos = (tmp_buf_pos + SAVE_BUF_SIZE - tmp_buf_size) % SAVE_BUF_SIZE;
                unsigned first_err_pos = buf_stat.errs[0].tmp_buf_pos;

                /* номер индекса относительно первой ошибки */
                int delta_idx = -(first_err_pos > cur_buf_pos ?
                                      first_err_pos - cur_buf_pos :
                                      first_err_pos - cur_buf_pos + tmp_buf_size);

                for (unsigned i = 0; i < tmp_buf_size; i++) {
                    dumpfile << "   " << setw(6) << delta_idx <<  " : " << setw(5) << tmp_buf[cur_buf_pos] << endl;
                    delta_idx++;
                    cur_buf_pos++;
                    if (cur_buf_pos == SAVE_BUF_SIZE)
                        cur_buf_pos = 0;
                }
                dumpfile.close();
                cout << "dump finished" << endl;
            }
        }
    }

    if (pI) {
        pI->CloseLDevice();
        pI->Release();
    }

    if(handle)
        dlclose(handle);
    return 0;
}
